import sys, os
import pyfem1d.pyfem1d_cmd as cmd_
import numpy as np
import collections
from pyfem1d.umat import *
from pyfem1d.load import *
from numpy import zeros, dtype, float64
from subprocess import call


class Analysis:
    def __init__(self):
        self.cmdShell = cmd_.Pyfem1dShell(self)

        self.input_file = None
        self.output_file = None
        self.logFile = None
        self.stress_file = None
        self.displacement_file = None
        self.plot_file = None
        self.interactive = False
        # self.gui = False
        self.o_node = None
        self.o_elem = None

        self.umat_dict = {}
        self.load_dict = {}

        self.current_umat_key = None
        self.current_load_key = None

        self.umat = None
        self.load = None

        self.bctype = None
        self.abspath = os.path.dirname(os.path.abspath(__file__))
        self.workingDirectory = None

        self.number_of_elements = None
        self.timestep = None
        self.maximum_time = None

    def header(self):
        header = ""

        header += format_if_not_none(" Input file    : ", self.input_file)
        header += format_if_not_none(" Output file   : ", self.output_file)
        header += format_if_not_none(" Stress file   : ", self.stress_file)
        header += format_if_not_none(" Disp. file    : ",
                                     self.displacement_file)

        header += format_if_not_none(" Number of Elements    nelem = ",
                                     self.number_of_elements)
        header += format_if_not_none(" Time step size           dt = ",
                                     self.timestep)
        header += format_if_not_none(" Duration               tmax = ",
                                     self.maximum_time)
        header += format_if_not_none(" Boundary Condition   bctype = ",
                                     self.bctype)

        header += format_if_not_none(" Umat : ", self.umat)
        header += format_if_not_none(" Load : ", self.load)

        return header

    def printHeader(self):
        """Prints program header with version"""
        print("pyfem1d - 1d finite elements for testing material formulations")
        #print(self.header())

    def add_umats(self, path):
        self.umat_dict.update(deploy_umats(path))

    def add_loads(self, path):
        self.load_dict.update(deploy_loads(path))

    def set_umat(self, key):
        self.umat = self.umat_dict[key]
        self.current_umat_key = key

    def set_umat_parameters(self, parameters):
        self.umat.parameter_values = parameters
        self.umat_dict[self.current_umat_key].parameter_values = parameters

    def set_load(self, key):
        self.load = self.load_dict[key]
        self.current_load_key = key

    def set_load_parameters(self, parameters):
        self.load.parameter_values = parameters
        self.load_dict[self.current_load_key].parameter_values = parameters

    def plotToWindow(self, stress_file=None):
        if not stress_file:
            stress_file = self.stress_file
        commands = ""
        commands += "set terminal X11 size 1300 400;"
        commands += "set key rmargin;"
        commands += "set multiplot;"
        commands += "set lmargin at screen 0.025;"
        commands += "set rmargin at screen 0.325;"
        commands += "set xlabel \"Time\";"
        commands += "set ylabel \"Strain\";"
        commands += "plot \"%s\" u 1:2 w l;" % stress_file
        commands += "set lmargin at screen 0.35;"
        commands += "set rmargin at screen 0.65;"
        commands += "set ylabel \"Stress\";"
        commands += "plot \"%s\" u 1:3 w l;" % stress_file
        commands += "set lmargin at screen 0.675;"
        commands += "set rmargin at screen 0.975;"
        commands += "set xlabel \"Strain\";"
        commands += "plot \"%s\" u 2:3 w l;" % stress_file
        commands += "unset multiplot;"

        call(["gnuplot", "-p", "-e", commands])

#set output "| ps2pdf -dCompatibilityLevel=1.4 -dPDFSETTINGS=/prepress - "+self.ofilebase+"_plot.pdf;\

    def plotPdf(self, stress_file, plot_file):
        # if not stress_file:
        #     stress_file = self.stress_file

        # if not plot_file:
        #     plot_file = self.plot_file

        print("Plotting to file %s" % plot_file)

        commands = ""
        commands += "set term pdf enhanced font \"Helvetica,10\";"
        commands += "set output \"%s\";" % plot_file
        commands += "set lmargin;"
        commands += "set rmargin;"
        commands += "set grid;"
        commands += "unset key;"
        commands += "set xlabel \"Time\";"
        commands += "set ylabel \"Strain\";"
        commands += "plot \"%s\" u 1:2 w l lt 1;" % stress_file
        commands += "set ylabel \"Stress\";"
        commands += "plot \"%s\" u 1:3 w l lt 1;" % stress_file
        commands += "set xlabel \"Strain\";"
        commands += "plot \"%s\" u 2:3 w l lt 1;" % stress_file

        call(["gnuplot", "-p", "-e", commands])

    def startShell(self):
        self.cmdShell.cmdloop()

    def setRuntimeParameters(self):
        self.o_node = self.number_of_elements + 1  #node to be written to the output
        self.o_elem = self.number_of_elements  #element to be written to the output
        self.number_of_elements = self.number_of_elements
        self.neq = self.number_of_elements + 1
        self.nnode = self.number_of_elements + 1
        self.t = 0

        if isinstance(self.bctype, collections.Iterable):
            self.bctype = 0

    def solve(self):
        '''Solution function'''

        if not self.umat:
            raise Exception("No umat function specified")

        if not self.load:
            raise Exception("No load function specified")

        if not self.maximum_time:
            raise Exception("No end time specified")

        if not self.number_of_elements:
            raise Exception("No number of elements specified")

        if not self.timestep:
            raise Exception("No timestep specified")

        self.setRuntimeParameters()
        output = open(self.output_file, 'w')

        #self.update()

        print(self.header())
        output.write(self.header())
        #self.printheader(ofile=output)

        neq = self.number_of_elements + 1
        #Initialize the system matrices
        self.init_system()
        #Set time to zero
        self.t = 0.
        #Generate nodes &  initialize solution vector
        dl = 1 / self.number_of_elements
        u = zeros((self.neq, 1))
        x = zeros((self.nnode, 1))
        #u= zeroflt(self.neq,1)
        #x= zeroflt(self.nnode,1)
        for i in range(int(self.nnode)):
            x[i] = (i - 1) * dl

        # open files for output at the selected node and element
        outd = open(self.displacement_file, 'w')
        outs = open(self.stress_file, 'w')

        # Simulation loop

        self.umat.initial_cond(self.number_of_elements)
        # if self.umat.initcond():
        #     self.umat.initcond()(self.number_of_elements)

        while self.t <= self.maximum_time + self.timestep:

            load = self.load.value(self.t)
            if self.verbose:
                print('   Compute solution at t= %6.4f, load= %6.4f' %
                      (self.t, load))

            # Newton iterations
            res = 1
            niter = 0
            nitermax = 50
            while res > 1.e-10 and niter < nitermax:
                #Calculate the residual vector and stiffness matrix
                #and update the history variables
                self.comp_stiffness(u)
                #impose boundary conditions
                if self.bctype == 0:
                    self.kg[:][0] = 0.
                    self.kg[0][:] = 0.
                    self.kg[0][0] = 1.
                    self.fg[-1] -= load
                    self.fg[0] = 0.
                elif self.bctype == 2:
                    du = load - u[-1]
                    for i in range(int(neq)):
                        self.fg[i] -= self.kg[i][-1] * du
                    self.kg[:][0] = 0.
                    self.kg[0][:] = 0.
                    self.kg[-1][:] = 0
                    self.kg[:][-1] = 0
                    self.kg[0][0] = 1.
                    self.kg[-1][neq - 1] = 1.
                    self.fg[0] = 0.
                    self.fg[-1] = -1 * du
                elif self.bctype == 1:
                    self.kg[:][0] = 0.
                    self.kg[0][:] = 0.
                    self.kg[0][0] = 1.
                    self.fg[1:-1] -= load * dl
                    self.fg[-1] = self.fg[-1] - load * dl / 2
                    self.fg[0] = 0
                else:
                    raise Exception('Error: Undefined bc type identifier: ' +
                                    self.bctype)
                #calculate the residual
                res = np.linalg.norm(self.fg, 2)
                if self.verbose:
                    print('  >> iter %2i, res = %10.5e' % (niter, res))
                #solve the system
                kginv = np.linalg.inv(self.kg)
                dg = np.dot(kginv, self.fg)
                #update nodal displacements
                u -= dg
                niter += 1
            # Print results
            #if self.verbose:
            output.write('\n Solution at t=%6.4f, load= %6.4f\n' %
                         (self.t, load))
            output.write('  Nodal solutions\n')
            output.write('  Node   x-coord.     u\n')
            for i in range(int(neq)):
                output.write(' %4i    %6.4e   %6.4e\n' % (i, x[i], u[i]))
            output.write('\n  Element solutions\n')
            output.write('  Element   Strain       Stress\n')
            for i in range(int(self.number_of_elements)):
                output.write(' %4i       %6.4e   %6.4e\n' %
                             (i, self.eps[i], self.sig[i]))
            outd.write(' %6.4e   %6.4e\n' % (self.t, u[self.o_node - 1]))
            outs.write(' %6.4e   %6.4e   %6.4e\n' %
                       (self.t, self.eps[self.o_elem - 1],
                        self.sig[self.o_elem - 1]))

            # Update History
            # if self.umat.update():
            # self.umat.update()()
            self.umat.update()

            self.t += self.timestep

        print('Finished solution')
        outd.close()
        outs.close()
        output.close()

    def init_system(self):
        self.fg = zeros((self.neq, 1))
        self.kg = zeros((self.neq, self.neq))
        self.eps = zeros((self.number_of_elements, 1))
        self.sig = zeros((self.number_of_elements, 1))
        self.aa = zeros((self.number_of_elements, 1))
        #self.fg=zeroflt(self.neq,1)
        #self.kg=zeroflt(self.neq,self.neq)
        #self.eps=zeroflt(self.number_of_elements,1)
        #self.sig=zeroflt(self.number_of_elements,1)
        #self.aa=zeroflt(self.number_of_elements,1)
        self.dl = 1 / float(self.number_of_elements)

    def comp_stiffness(self, u):
        '''Compute stiffness matrix and residual vector,
        and update history variables'''
        offs = 0
        self.fg[:] = 0.
        self.kg[:][:] = 0.
        dN = zeros((2, 1))
        #dN=zeroflt(2,1)
        # Assembly: loop over elements
        for n in range(int(self.number_of_elements)):
            dN[0] = -1 / self.dl
            dN[1] = -1 * dN[0]
            #compute element strain
            epsl = (u[n + 1] - u[n]) / self.dl
            # Calculate the stress, modulus and update history at the Gauss point

            sigl, aal = self.umat.stress_tangent(self.timestep, n, epsl)

            #store the stresses and moduli
            self.eps[n] = epsl
            self.sig[n] = sigl
            self.aa[n] = aal
            #loop over element dofs: element stiffness and internal force vector
            #fe = zeroflt(2,1)
            #ke = zeroflt(2,2)
            fe = zeros((2, 1))
            ke = zeros((2, 2))
            for i in range(2):
                fe[i] += dN[i] * sigl * self.dl
                #\int B^{T}\sigma dV
                for j in range(2):
                    ke[i][j] += dN[i] * aal * dN[
                        j] * self.dl  #\int B^{T} E B dV

            #assemble global matrices
            for i in range(2):
                self.fg[i + offs] += fe[i]
                for j in range(2):
                    self.kg[i + offs][j + offs] += ke[i][j]
            offs += 1


def format_if_not_none(prefix, var):
    return prefix + (str(var) if var else "None") + "\n"



from pyfem1d import helper

class Umat:
    parameter_values = []
    parameter_names = []

    def __init__(self):
        pass

    def initial_cond(self, n_elem):
        pass

    def update(self):
        pass

    def stress_tangent(self, dt, n, eps):
        raise Exception("Stress-tangent function of umat not defined")

    def set_parameter_values(self, values):
        pass

    def __repr__(self):
        result = ""
        result += type(self).__name__ + "("
        for n, (name, val) in enumerate(zip(self.parameter_names, self.parameter_values)):
            result += name+"="+str(val)
            if n != len(self.parameter_names)-1:
                result += ", "
        result += ")"
        return result

def deploy_umats(path):
    return helper.deploy_objects(path, Umat)



from collections import OrderedDict
import inspect


def deploy_objects(path, class_):
    local_ret = globals()
    # local_ret = {}
    global_ret = globals()
    with open(path) as f:
        code = compile(f.read(), path, 'exec')
        exec(code, global_ret, local_ret)

    # globals().update(global_ret)

    # globals().update(local_ret)

    result = {}
    try:
        for key, i in local_ret.items():
            # print(key, i.__class__)
            if inspect.isclass(i):
                if issubclass(i, class_) and key != class_.__name__:
                    # print(key, i)
                    result[key] = i()
    except Exception as e:
        raise Exception("There was an error in the config %s:\n%s" %
                        (path, str(e)))

    # Sort the result dict according to the keys alphabetically
    result = OrderedDict(sorted(result.items(), key=lambda t: t[0]))
    return result



from pyfem1d import helper

class Load:
    parameter_names = []
    parameter_values = []

    def __init__(self):
        pass


    def __repr__(self):
        result = ""
        result += type(self).__name__ + "("
        for n, (name, val) in enumerate(zip(self.parameter_names, self.parameter_values)):
            result += name+"="+str(val)
            if n != len(self.parameter_names)-1:
                result += ", "
        result += ")"
        return result


def deploy_loads(path):
    return helper.deploy_objects(path, Load)



from pyfem1d.load import Load
from math import sin, pi


class Triangle(Load):
    parameter_values = [25, 0.008]
    parameter_names = ['loading_per', 'magnitude']

    def value(self, t):
        tper = self.parameter_values[0]
        magn = self.parameter_values[1]
        tloc = t
        while tloc > tper:
            tloc = tloc - tper
        qper = tper / 4.
        if tloc <= qper:
            load = tloc
        elif tloc > qper and tloc <= 2. * qper:
            load = qper - (tloc - qper)
        elif tloc > 2. * qper and tloc <= 3. * qper:
            load = -1. * (tloc - 2. * qper)
        elif tloc > 3. * qper and tloc <= tper:
            load = -1. * qper + (tloc - 3. * qper)
        return load * magn


class Sine(Load):
    parameter_values = [25., 0.008]
    parameter_names = ['loading_per', 'magnitude']

    def value(self, t):
        return self.parameter_values[1] * sin(t / self.parameter_values[0] *
                                              2. * pi)


class Creep(Load):
    parameter_values = [0, 10, 0.008]
    parameter_names = ['start', 'end', 'magnitude']

    def value(t):
        if t <= self.parameter_values[1] and t >= self.parameter_values[0]:
            return self.parameter_values[2]
        else:
            return 0


class Heaviside(Load):
    parameter_values = [0.008]
    parameter_names = ['magnitude']

    def value(t):
        magn = self.parameter_values[0]
        return magn


class Proportional(Load):
    parameter_values = [0.008]
    parameter_names = ['magnitude']

    def value(t):
        magn = self.parameter_values[0]
        return magn * t



from pyfem1d import Analysis
from multiprocessing import Process
import argparse
import os
import ntpath

import pyfem1d.umat_defaults
import pyfem1d.load_defaults

# abspath = os.path.dirname(os.path.abspath(__file__))


class FullPaths(argparse.Action):
    """Expand user- and relative-paths"""

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest,
                os.path.abspath(os.path.expanduser(values)))


def is_dir(dirname):
    """Checks if a path is an actual directory"""
    if not os.path.isdir(dirname):
        msg = "{0} is not a directory".format(dirname)
        raise argparse.ArgumentTypeError(msg)
    else:
        return dirname


parser = argparse.ArgumentParser()
parser.add_argument("input",
                    metavar="input_file",
                    nargs="?",
                    help="input file",
                    type=argparse.FileType("r"))
# parser.add_argument("-o",
#                     "--output-file",
#                     default="default_out.dat",
#                     type=argparse.FileType("w"))
#parser.add_argument("-l", "--log-file", default="default.log",
#type=argparse.FileType("w"))
# parser.add_argument("-d",
#                     "--displacement-file",
#                     default="default_disp.dat",
#                     type=argparse.FileType("w"))
# parser.add_argument("-u",
#                     "--stress-file",
#                     default="default_stre.dat",
#                     type=argparse.FileType("w"))
# parser.add_argument("-p",
#                     "--plot-file",
#                     default="default.ps",
#                     type=argparse.FileType("w"))
# parser.add_argument("-n",
#                     "--number-of-elements",
#                     default="10",
#                     type=int)

# parser.add_argument("-t", "--timestep", default="0.1", type=float)
# parser.add_argument("-m", "--maximum-time", default="25", type=float)
# parser.add_argument("-g", "--gui", action="store_true", help="Start the graphical user interface")
parser.add_argument("-v", "--verbose", action="store_true")

# parser.add_argument("-s", "--silent", action="store_true")
# parser.add_argument("-i", "--interactive", action="store_true")


def __main__():
    args = parser.parse_args()

    an = Analysis()

    # an.silent = args.silent
    an.verbose = args.verbose
    # an.interactive = args.interactive
    # an.gui = args.gui

    # input file
    if args.input:
        an.input_file = os.path.abspath(args.input.name)
        an.workingDirectory = os.path.dirname(os.path.abspath(args.input.name))
    else:
        an.workingDirectory = os.getcwd()

    #an.logFile = os.path.abspath(args.log_file.name)
    # an.output_file = os.path.abspath(args.output_file.name)
    # an.stress_file = os.path.abspath(args.stress_file.name)
    # an.displacement_file = os.path.abspath(args.displacement_file.name)
    # an.plotFile = os.path.abspath(args.plot_file.name)

    an.output_file = "default_out.dat"
    an.stress_file = "default_stre.dat"
    an.displacement_file = "default_disp.dat"

    an.add_umats(os.path.abspath(pyfem1d.umat_defaults.__file__))
    an.add_loads(os.path.abspath(pyfem1d.load_defaults.__file__))

    an.printHeader()

    # Execute the input file if given
    if an.input_file:
        an.cmdShell.execFile(an.input_file)
    else:
        an.startShell()


if __name__ == "__main__":
    __main__()



import cmd
import sys
import os
# import pyfem1d.tkgui

class Pyfem1dShell(cmd.Cmd):
    intro = 'Type help or ? to list commands.\n'
    prompt = 'pyfem1d>> '
    file = None

    def __init__(self,parent):
        cmd.Cmd.__init__(self)
        #super(Pyfem1dShell, self).__init__()
        self.analysis = parent

    def do_verbose(self, arg):
        '''Enable verbose output'''
        self.analysis.verbose = True
        #print(arg)

    # def do_gui(self, arg):
    #     '''Start graphical user interface'''
    #     pyfem1d.tkgui.startGui(self.analysis)

    def do_dt(self, arg):
        '''Set timestep size: dt value'''
        val, errormsg = parse_line(arg, type = float, n_args = 1)
        if not errormsg:
            self.analysis.timestep = val[0]
        else:
            print(errormsg)

    def do_nelem(self, arg):
        '''Set total number of elements with nelem(value)'''
        val, errormsg = parse_line(arg, type = int, n_args = 1)
        if not errormsg:
            self.analysis.number_of_elements = val[0]
        else:
            print(errormsg)

    def do_tmax(self, arg):
        '''Set maximum time tmax(value)'''
        val, errormsg = parse_line(arg, type = float, n_args = 1)
        if not errormsg:
            self.analysis.maximum_time = val[0]
        else:
            print(errormsg)

    def do_bctype(self, arg):
        '''Set bc type with bctype(val), can be 0,1,2'''
        val, errormsg = parse_line(arg, type = int, n_args = 1)
        if not errormsg:
            if val[0] == 0 or val[0] == 1 or val[0] == 2:
                self.analysis.bctype = val[0]
            else:
                print("Error: bctype can only be one of 0, 1 and 2")
        else:
            print(errormsg)


    def do_solve(self, arg):
        '''Start solution with solve()'''
        try:
            self.analysis.solve()
        except Exception as e:
            print(e)

    def do_plot(self, arg):
        '''Plot the output stress file using gnuplot. optional argument: stress_file'''
        val, errormsg = parse_line(arg, type = str)
        if len(val) == 0:
            stress_file = self.analysis.stress_file
        elif len(val) == 1:
            stress_file = val[0]
        else:
            print("Error: plot cannot receive more than 1 arguments")

        self.analysis.plotToWindow(stress_file=stress_file)

    def do_plotpdf(self, arg):
        '''Plot the output stress file to a pdf file using gnuplot. optional arguments: plot_file stress_file'''
        val, errormsg = parse_line(arg, type = str)
        if len(val) == 0:
            plot_file = self.analysis.plot_file
            stress_file = self.analysis.stress_file
        elif len(val) == 1:
            plot_file = val[0]
            stress_file = self.analysis.stress_file
        elif len(val) == 2:
            plot_file = val[0]
            stress_file = val[1]
        else:
            print("Error: plot cannot receive more than 2 arguments")

        self.analysis.plotPdf(stress_file=stress_file, plot_file=plot_file)

    def do_pwd(self, arg):
        'Prints current working directory'
        print(self.analysis.workingDirectory)

    #def do_cd(self, arg):
        #'Changes current working directory'

    def do_addumats(self, arg):
        '''Add umat material: addumat exampleMaterial.py'''
        val, errormsg = parse_line(arg, type = str, n_args = 1)
        if not errormsg:
            try:
                path = os.path.join(self.analysis.workingDirectory, val[0])
                self.analysis.add_umats(path)
            except Exception as e:
                print(e)
        else:
            print(errormsg)

    def do_addloads(self, arg):
        '''Add loading function: addloading exampleLoading.py'''
        val, errormsg = parse_line(arg, type = str, n_args = 1)
        if not errormsg:
            try:
                path = os.path.join(self.analysis.workingDirectory, val[0])
                self.analysis.add_loads(path)
            except Exception as e:
                print(e)
        else:
            print(errormsg)


    def do_load(self, arg):
        '''Set load function: load triangle '''
        val, errormsg = parse_line(arg, type = str, n_args = "1+")
        if not errormsg:
            self.analysis.set_load(val[0])
            if len(val) > 1:
                parameters = tuple(map(float, val[1:]))
                self.analysis.set_load_parameters(parameters)
        else:
            print(errormsg)

    def do_listumats(self, arg):
        '''List all umats with the parameters and their values'''
        for key, umat in self.analysis.umat_dict.items():
            print(umat)

    def do_listloads(self, arg):
        '''List all loading functions with the parameters and their values'''
        for key, load in self.analysis.load_dict.items():
            print(load)

    def do_vars(self, arg):
        '''List all variable values and current functions'''
        print(self.analysis.header())


    def do_umat(self, arg):
        '''Set umat function: umat maxwell '''
        val, errormsg = parse_line(arg, type = str, n_args = "1+")

        if not errormsg:
            try:
                self.analysis.set_umat(val[0])
                if len(val) > 1:
                    parameters = tuple(map(float, val[1:]))
                    self.analysis.set_umat_parameters(parameters)
            except Exception as e:
                print(e)
        else:
            print(errormsg)

    def do_run(self, arg):
        val, errormsg = parse_line(arg, type = str, n_args = 1)
        if not errormsg:
            try:
                self.analysis.cmdShell.execFile(val[0])
            except Exception as e:
                print(e)
        else:
            print(errormsg)

    def do_q(self, arg):
        '''End session'''
        self.close()
        return True

    def do_EOF(self,arg):
        '''End session'''
        self.close()
        return True

    # ----- record and playback -----
    #def do_record(self, arg):
        #'Save future commands to filename:  RECORD rose.cmd'
        #self.file = open(arg, 'w')
    #def do_playback(self, arg):
        #'Playback commands from a file:  PLAYBACK rose.cmd'
        #self.close()
        #with open(arg) as f:
            #self.cmdqueue.extend(f.read().splitlines())

    def precmd(self, line):
        # line = line.lower()
        if self.file and 'playback' not in line:
            print(line, file=self.file)
        return line

    def close(self):
        if self.file:
            self.file.close()
            self.file = None

    def execFile(self, filename):
        f = open( filename, "r" )
        array = []
        for line in f:
            array.append( line )
        f.close()
        for i,j in enumerate(array):
            try:
                line = cleanLine(j)
                if(line):
                    #print("Executing: "+line)
                    self.onecmd(line)
            except Exception as err:
                print(err)


def cleanLine(line):
    return line.split('#')[0].strip()

def parse_line(arg, n_args = None, type=int):
    'Convert a series of zero or more numbers to an argument tuple'
    result = tuple(map(type, arg.split()))
    if isinstance(n_args, str):
        if n_args[-1] == "+":
            min_limit = int(n_args[:-1])
            if len(result) < min_limit:
                errormsg = "expected minimum "+str(min_limit)+" arguments, received "+str(len(result))
            else:
                errormsg = None
    elif isinstance(n_args, int):
        if n_args and len(result) != n_args:
            errormsg = "expected "+str(n_args)+" arguments, received "+str(len(result))
        else:
            errormsg = None
    else:
        errormsg = None
    return result, errormsg




from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))


setup(
    name="pyfem1d",

    version="0.0.2.post1",

    description="1d finite elements for testing material formulations",

    author="H. Onur Solmaz",

    author_email="hos@solmaz.co",

    packages=find_packages(exclude=["contrib", "docs", "tests"]),

    url = "https://github.com/hos/pyfem1d",

    download_url = "https://github.com/hos/pyfem1d/tarball/0.0.2.post1",

    keywords = ["FEM", "materials", "mechanics"],

    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage"],
    },

    # package_data={
    #     "sample": ["package_data.dat"],
    # },

    # data_files=[("my_data")],

    install_requires={
    },

    entry_points={
        "console_scripts": [
            "pyfem1d=pyfem1d.main:__main__",
        ],
    },
)






import numpy as np
from pyfem1d.umat import Umat


class LinearElastic(Umat):
    parameter_values = [100]
    parameter_names = ['elastic_mod']

    def stress_tangent(self, dt, n, eps):
        #Get the material variables
        E = self.parameter_values[0]
        #Calculate the stress and consistent modulus
        sigl = E * eps  #linear elasticity
        aal = E
        return sigl, aal


class NonlinearElastic(Umat):
    # parameter_values = [100, 0.03]
    parameter_values = [100, 0.0001]
    parameter_names = ['elastic_mod', 'sat_rate']

    def stress_tangent(self, dt, n, eps):
        E = self.parameter_values[0]
        eta = self.parameter_values[1]

        #Calculate the stress and consistent modulus
        sigl = E * eta * (1 - np.exp(-1 * eps / eta))  #nonlin elasticity
        aal = E * np.exp(-1 * eps / eta)
        return sigl, aal


class Maxwell(Umat):

    parameter_values = [100, 1500]
    parameter_names = ['elastic_mod', 'viscosity']

    def __init__(self):
        self.h1 = []
        self.hn = []

    def initial_cond(self, asdf):
        self.h1 = np.zeros((asdf, 1), dtype=np.float64)
        self.hn = np.zeros((asdf, 1), dtype=np.float64)
        self.nelem = asdf

    def update(self):
        temp = self.h1[:]
        self.h1 = self.hn[:]
        self.hn = temp[:]

    def stress_tangent(self, dt, n, eps):
        #Get the material variables
        E = self.parameter_values[0]
        eta = self.parameter_values[1]
        alphan = self.hn[n]
        #Calculate the stress and consistent modulus
        alpha = (alphan + eps * dt * E / eta) / (1 + dt * E / eta)
        sigl = E * (eps - alpha)
        aal = E / (1 + dt * E / eta)

        self.h1[n] = alpha

        return sigl, aal


class StandardLinearSolid(Umat):
    parameter_values = [1000, 4000, 20000]
    parameter_names = ['E0', 'E1', 'viscosity']

    def __init__(self):
        self.h1 = []
        self.hn = []

    def initial_cond(self, asdf):
        self.h1 = np.zeros((asdf, 1), dtype=np.float64)
        self.hn = np.zeros((asdf, 1), dtype=np.float64)
        self.nelem = asdf

    def update(self):
        global h1, hn
        temp = self.h1[:]
        self.h1 = self.hn[:]
        self.hn = temp[:]

    def stress_tangent(self, dt, n, eps):
        #Get the material variables
        E0 = self.parameter_values[0]
        E1 = self.parameter_values[1]
        eta = self.parameter_values[2]
        alphan = self.hn[n]

        #Calculate the stress and consistent modulus
        alpha = (alphan + eps * dt * E1 / eta) / (1 + dt * E1 / eta)

        sigl = E0 * eps + E1 * (eps - alpha)
        aal = E0 + E1 / (1 + dt * E1 / eta)  #
        #Update history
        self.h1[n] = alpha

        return sigl, aal


class StandardLinearSolid2(Umat):

    parameter_values = [1000, 4000, 20000]
    parameter_names = ['E0', 'E1', 'viscosity']

    #step size for the five-point stencil
    fs = 0.001
    tol = 1e-10
    offset = 0.1

    def __init__(self):
        self.h1 = []
        self.hn = []

    def initial_cond(self, asdf):
        self.h1 = np.zeros((asdf, 1), dtype=np.float64)
        self.hn = np.zeros((asdf, 1), dtype=np.float64)
        self.nelem = asdf

    def update(self, ):
        temp = self.h1[:]
        self.h1 = self.hn[:]
        self.hn = temp[:]

    # Constitutive equations

    def sigel1(self, strain):  #elastic stress
        result = strain * self.parameter_values[1]
        return result

    def sigel0(self, strain):  #elastic stress
        result = strain * self.parameter_values[0]
        return result

    def epsvisdot(self, stress):  # viscous strain derivative
        result = stress / self.parameter_values[2]
        return result

    def sig(self, eps, alpha):  #total stress
        result = self.sigel0(eps) + self.sigel1(eps - alpha)
        return result

    # Secand iteration

    def nextstep(self, hist, dt, eps):
        dummy = [hist - self.offset, hist]
        #using the secant method
        #print '     ++++++++++++++++'
        while True:
            temp2 = dummy[1] - self.residual(dummy[1], hist, dt, eps) * (
                dummy[1] - dummy[0]) / (self.residual(dummy[1], hist, dt, eps)
                                        - self.residual(dummy[0], hist, dt,
                                                        eps))
            dummy[0] = dummy[1]
            dummy[1] = temp2
            err = abs(dummy[0] - dummy[1])

            #print '     >>>>> Secant_err: %10.5e' %(err)
            if err < self.tol: break
        return dummy[1]

    # RESIDUAL
    def residual(self, next_, now, dt, eps):
        result = next_ - now - dt * self.epsvisdot(self.sigel1(eps - next_))
        return result

    # Five point stencil

    #def fivepoint(f,*p):
    #return (-1*f(p[0]+2*fs,p[1:])+8*f(p[0]+fs,p[1:])-8*f(p[0]-fs,p[1:])+f(p[0]-2*fs,p[1:]))/(12*fs)

    def stress_tangent(self, dt, n, eps):
        alphan = self.hn[n]

        #CALCULATE THE NEXT VISCOUS STRAIN
        alpha = self.nextstep(alphan, dt, eps)

        #calculate the stress and the consistent modulus
        sigl = self.sig(eps, alpha)
        #aal  = fivepoint(sig,eps,alpha)
        aal = (-1 * self.sig(eps + 2 * self.fs, alpha) + 8 * self.sig(
            eps + self.fs, alpha) - 8 * self.sig(eps - self.fs, alpha) +
               self.sig(eps - 2 * self.fs, alpha)) / (12 * self.fs)
        #Update history
        self.h1[n] = alpha

        return sigl, aal
